import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.xml.sax.ext.LexicalHandler;
import java.io.StringReader;

/*

Modification History
--------------------

Version 1.0
	This example shows how the elements of XML data are read into a program

Version 2.0
	This example shows how the errors in XML data are trapped
    Note that the application gets the callbacks till the parsing error is found

Version 3.0

	This example demonstrates how attributes are read in
	Two access mechanisms demonstrated
	1.by attribute index;
    2.by qualified (prefixed) name.

Version 4.0
	This example demonstrates how PIs are handled

	Note: The first line in the CarParts.xml is a PI (XML document declaration)
	that identifies XML document and its encoding scheme.

	The content is all after the first XML document declaration.

Version 2A01.0
	Handling characters inside an element

Version 2A02.0
	Adding DTD - Note the impact of adding DTD on the characters
	Still using non-validating mode - Note the parsing does not catch the error in XML w.r.t DTD
	Uncommenting the ignorableWhitespace will get the ignored characters back

Version 2A03.0
	Set DTD based Validation On - xml error gets caught

Version 0401
	Setting the DocumentLocation and using the Locator object to identify the document and even locations

Version 0402
	Implementing the Lexicalhandler

Version 0403
	Implementing Entity Resolver
*/

public class MyXMLHandler extends DefaultHandler implements LexicalHandler{

	Locator l1;
	public void setDocumentLocator(Locator l)
	{
		System.out.println("Locator :" + l.getSystemId() + "\n");
		l1 = l;
	}
	public void startDocument()
    {
        System.out.println("\n Start Document: -----Reading the document CarParts.xml with MyXMLHandler------\n");
    }

    public void startElement(String namespaceURI, String localName,
                             String qualifiedName, Attributes elementAttributes)
    {
        System.out.println("Location of event at line number :" + l1.getLineNumber() + "\n");
        System.out.println("Start Element-> "+qualifiedName);

								/* Get the index of price attributes*/
								int indexOfPrice=elementAttributes.getIndex("price");

								/* If a price attribute does not exist then all the attributes are printed out*/
								/* Note that in the sample the attribute name is case sensitive so all attributes
									of carstereo get printed out*/
								if(-1==indexOfPrice)
									printAllAttributes(elementAttributes);
								else
			System.out.println("\tPrice= "+elementAttributes.getValue("price"));
    }

    public void endElement(String namespaceURI, String localName,
                             String qualifiedName)
    {
        System.out.println("End Element-> "+qualifiedName);
    }

	public void characters(char[] ch, int start, int length)
	{
		System.out.println("Characters: " + new String(ch,start,length));
	}

    public void endDocument()
    {
        System.out.println("\n End Document: ----------------Finished Reading the document---------------------\n");
    }

	public void printAllAttributes(Attributes elementAttributes)
		{
			System.out.println("\tTotal Number of Attributes: "+elementAttributes.getLength());
			for(int i=0;i<elementAttributes.getLength();i++)
			{
				System.out.println("\t\tAttribute: "+elementAttributes.getQName(i)+
													    " = "+elementAttributes.getValue(i));
			}
	}


	//Lexical Handler Methods

	public void startCDATA()
		                throws SAXException
		{

			System.out.println("\nStarting CDATA Section\n");
		}


		public void endCDATA()
		              throws SAXException
		{

			System.out.println("\nEnding CDATA Section\n");
		}

		public void comment(char[] ch,
		                    int start,
		                    int length)
		             throws SAXException
		{

			String commentText = new String(ch,start,length);
			System.out.println("Comment : " + commentText + "\n");

		}
	public void startDTD(java.lang.String name,
	                     java.lang.String publicId,
	                     java.lang.String systemId)
	              throws SAXException
	 {

		 System.out.println("Starting DTD :" + systemId);

		}

	 public void endDTD()
	             throws SAXException
	{
			System.out.println("Ending DTD ");

	}

	public void startEntity(java.lang.String name)
	                 throws SAXException
	{

		System.out.println("Starting entity :" + name);
	}

	public void endEntity(java.lang.String name)
	               throws SAXException
	{

	System.out.println("Ending entity :" + name);
	}





    static public void main(String[] args) throws Exception {

        System.out.println("\nVersion 0403.0 of MyXMLHandler in example0403\n");

		/*Create a SAX Parser Factory*/
        SAXParserFactory parseFactory = SAXParserFactory.newInstance();

		/*Set to generate Validating SAX Parser */
		parseFactory.setValidating(true);

		/*Obtain a validating SAX Parser */
        SAXParser saxParser = parseFactory.newSAXParser();

		/*XML Reader is the interface for reading an XML document using callbacks*/
        XMLReader xmlReader = saxParser.getXMLReader();
        xmlReader.setProperty("http://xml.org/sax/properties/lexical-handler",new MyXMLHandler());

		/*Override the entity resolver*/
		xmlReader.setEntityResolver(new MyEntityResolver());

		/*set the error handler*/
		xmlReader.setErrorHandler(new MyErrorHandler());

		/*Attach ContentHandler - the callbacks like startDocument,startElement etc. are o
		  overridden by the setContentHandler to trap them into user code*/
        xmlReader.setContentHandler(new MyXMLHandler());

		try{
			/*Parse an XML document - the document is read and overridden callbacks in the MyXMLHandler are invoked*/
	        xmlReader.parse("CarParts.xml");
		}
		catch (SAXParseException saxException) {
					/* If there are errors in XML data are trapped and location is displayed*/
					System.out.println("\n\nError in CarParts.xml at line:"+saxException.getLineNumber()+"("+saxException.getColumnNumber()+")\n");
					System.out.println(saxException.toString());
				}
		catch (SAXException saxEx) {
					/* If there are errors in XML data, the detailed message of the exception is displayed*/
					System.out.println(saxEx.getMessage());
				}
    }


    static class MyErrorHandler implements ErrorHandler
	{

        public void fatalError(SAXParseException saxException)
		{
            System.out.println("Fatal Error occurred "+ saxException);
        }

        public void error(SAXParseException saxException)
		{
            System.out.println("Error occurred "+ saxException);
        }

        public void warning(SAXParseException saxException)
		{
            System.out.println("warning occurred "+ saxException);
        }
    }


    static class MyEntityResolver implements EntityResolver
		{

			static String xmlDTD =

				"<!ENTITY  companyname \"Heaven Car Parts (TM)\">"+
				"<!ENTITY  companyweb \"http://carpartsheaven.com\">"+
				"<!ELEMENT carparts (supplier,engines,carbodies,wheels,carstereos,forCDATA)>"+
				"<!ELEMENT engines (engine+)>"+
				"<!ELEMENT carbodies (carbody+)>"+
				"<!ELEMENT wheels (wheel+)>"+
				"<!ELEMENT carstereos (carstereo+)>"+
				"<!ELEMENT forCDATA ANY>"+
				"<!ELEMENT supplier (#PCDATA)>"+
				"<!ATTLIST supplier"+
				"			name CDATA #REQUIRED"+
				"			URL CDATA #REQUIRED"+
				">"+
				""+
				"<!ELEMENT engine (#PCDATA)*>"+
				"<!ATTLIST engine"+
				"			id CDATA #REQUIRED"+
				"			type CDATA #REQUIRED"+
				"			capacity (1000 | 2000 | 2500 ) #REQUIRED"+
				"			price CDATA #IMPLIED"+
				"			text CDATA #IMPLIED"+
				">"+
				"<!ELEMENT carbody (#PCDATA)*>"+
				"<!ATTLIST carbody"+
				"			id CDATA #REQUIRED"+
				"			type CDATA #REQUIRED"+
				"			color CDATA #REQUIRED"+
				">"+
				"<!ELEMENT wheel (#PCDATA)*>"+
				"<!ATTLIST wheel"+
				"			id CDATA #REQUIRED"+
				"			type CDATA #REQUIRED"+
				"			price CDATA #IMPLIED"+
				"			size  (X | Y | Z) #IMPLIED"+
				">"+
				"<!ELEMENT carstereo (#PCDATA)*>"+
				"<!ATTLIST carstereo"+
				"			id CDATA #REQUIRED"+
				"			manufacturer CDATA #REQUIRED"+
				"			model CDATA #REQUIRED"+
				"			Price CDATA #REQUIRED"+
				">";


			public InputSource resolveEntity(java.lang.String publicId, java.lang.String systemId)
			{
				if (systemId.equals("http://www.carpartsheaven.com/xml/dtds/carparts.dtd"))
				{
					System.out.println("Redirecting to local DTD for "+systemId);
					return new InputSource(new StringReader(xmlDTD));
				}
				else
				{
		    		return null;
				}
			}
		}


}
